<?php
require_once "OWLClass.php";
require_once "OWLOntology.php";
require_once "OWLTag.php";



/**
 * Interpretador de ontologias OWL.
 * @author Joel Azevedo, n39083
 * @author Diogo Ribeiro, n41567 
 * @version 1.0
 */
class OWLReader
{
    public static $DEBUG = false; // Mostrar informaes de debugging.
    
    private $errorCode;          // Indica o cdigo de erro. Se 0, no ocorreu erro.
    private $errorLine;          // Indica a linha onde ocorreu erro. Se -1, no ocorreu erro.
    
    

    /**
     * Construtor.
     */
    public function OWLReader()
    {
        $this->errorCode = 0;
        $this->errorLine = -1;
    }
    
    
    
    /**
     * L uma ontologia a partir de ficheiro.
     * @param string $path O caminho do ficheiro da ontologia a carregar.
     * @param OWLOntology ontology O apontador para a topologia a carregar.
     * @return <code>true</code> caso o ficheiro tenha sido encontrado e lido
     * com sucesso; <code>false</code> em caso contrrio.
     */
    public function loadFromFile($path, &$ontology)      
    {
        // Inicializar variveis:
        $namespace = $path;
        $rootClass = null;
        $ontology->setPath($path);
        
        // Abrir ficheiro:
        $file = @fopen($path, 'r');
        if(!$file) return false;
        
        // Inicializar parser:
        $parser = xml_parser_create_ns();
        $rootTag = new OWLTag($ontology, null);
        
        xml_parser_set_option($parser, XML_OPTION_CASE_FOLDING, false);
        xml_set_object($parser, &$rootTag);
        xml_set_element_handler($parser, "openTag", "closeTag");
        xml_set_character_data_handler($parser, "data");
        xml_set_start_namespace_decl_handler($parser, "nsDeclTag");
        
        // Efectuar parsing:
        if(OWLReader::$DEBUG) print "<pre>";
        while($data = fread($file, 4096))
        {
            xml_parse($parser, utf8_encode($data), feof($file));
            $this->errorCode = xml_get_error_code($parser);
            $this->errorLine = xml_get_current_line_number($parser) - 1;
        }   
        if(OWLReader::$DEBUG) print "</pre>";

        // Fechar ficheiro:
        fclose($file);
        
        return true;
    }
    
    
    
    /**
     * Devolve a descrio do erro.
     * @return string A string com a descrio do erro, ou <code>null</code>
     * caso no tenha ocorrido erro ao ler o ficheiro.
     */
    public function getErrorDescription()
    {
        if($this->errorCode == 0) return null;
        else return xml_error_string($this->errorCode);
    }
    
    
    
    /**
     * Devolve a linha onde ocorreu o erro.
     * @return integer A linha onde ocorreu o erro de leitura, ou
     * <code>-1</code> caso no tenha ocorrido erro ao ler o ficheiro.
     */
    public function getErrorLine()
    {
        if($this->errorCode == 0) return -1;
        else return $this->errorLine;
    }
    
    
    
    /**
     * Inferir hierarquia de classes de uma topologia.
     * NOTA: Esta classe existe porque no  possvel inferir as subclasses
     * correctamente antes de todas serem conhecidas. Os ficheiros OWL s
     * especificam superclasses. Alm disso, pode nem sempre ser necessrio
     * inferir subclasses, pelo que fica como opo ao utilizador.
     * @param OWLOntology $ontology Um apontador para a ontologia.
     */
    public function inferClassHierarchy(&$ontology)
    {
        // Percorrer todas as classes existentes:
        foreach($ontology->getClassURIs() as $classURI)
        {
            // Verificar se esta classe tem superclasses definidas:
            $superClasses = $ontology->_superClasses[$classURI];
            if(count($superClasses) > 0)
            {
                // Obter o elemento actual:
                $subClass = $ontology->getClass($classURI);
                if($subClass == null) continue;
                
                // Percorrer a ontologia:
                foreach($ontology->_superClasses[$classURI] as $superClassURI)
                {
                    // Efectuar a ligao, caso no exista:
                    if(!in_array($superClassURI, $subClass->getSuperClassURIs()))
                        $subClass->addSuperClassURI($superClassURI);
                    
                    // Obter este elemento:
                    $superClass = $ontology->getClass($superClassURI);
                    if($superClass == null) continue;
                    
                    // Efectuar a ligao inversa, caso no exista:
                    if(!in_array($classURI, $superClass->getSubClassURIs()))
                        $superClass->addSubClassURI($classURI);
                }
            }
            else
            {
                // No h superclasses definidas. Caso esta no seja "Thing",
                // ligar esta classe actual a "Thing":
                if(strcmp($classURI, $ontology->getThingClassURI()) == 0)
                    continue;
                
                // Estrutura OWL2 (sem "Thing"). Defini-la como superclasse
                $subClass = $ontology->getClass($classURI);
                if($subClass == null) continue;
                
                // Efectuar a ligao, caso no exista:
                if(!in_array($ontology->getThingClassURI(), $subClass->getSuperClassURIs()))
                    $subClass->addSuperClassURI($ontology->getThingClassURI());
                
                // Efectuar a ligao inversa, caso no exista:
                $thingClass = $ontology->getClass($ontology->getThingClassURI());
                if($thingClass == null) continue;
                
                if(!in_array($classURI, $thingClass->getSubClassURIs()))
                    $thingClass->addSubClassURI($classURI);
            }
        }
    }
    
    
    
    /**
     * Inferir subclasses de uma topologia.
     * NOTA: Esta classe existe porque no  possvel inferir as subclasses
     * correctamente antes de todas serem conhecidas. Os ficheiros OWL s
     * especificam superclasses. Alm disso, pode nem sempre ser necessrio
     * inferir subclasses, pelo que fica como opo ao utilizador.
     * @param OWLOntology $ontology Um apontador para a ontologia.
     */
    public function inferClassMembers(&$ontology)
    {
        // Percorrer todos os indivduos existentes:
        foreach($ontology->getIndividualURIs() as $individualURI)
        {
            // Obter o objecto deste indivduo:
            $individual = $ontology->getIndividual($individualURI);
            if($individual == null) continue;
            
            // Verificar se este indivduo tem classes-tipo:
            if(count($ontology->_members[$individualURI]) == 0) continue;
            
            // Obter as classes-tipo deste indivduo:
            foreach($ontology->_members[$individualURI] as $classURI)
            {
                // Estabelecer a associao directa:
                $individual->addTypeClass($classURI);
                
                // Obter o objecto desta classe:
                $class = $ontology->getClass($classURI);
                if($class == null) continue;
                
                // Estabelecer a associao inversa:
                $class->addMemberURI($individualURI);
            }
        }
    }
    
    
    
    /**
     * Inferir a hierarquia das propriedades de objecto de uma topologia.
     * @param OWLOntology $ontology Um apontador para a ontologia.
     */
    public function inferOPHierarchy(&$ontology)
    {
        // Percorrer todas as propriedades de objecto existentes:
        foreach($ontology->getObjectPropertyURIs() as $propertyURI)
        {
            // Obter o objecto da propriedade:
            $property = $ontology->getObjectProperty($propertyURI);
            if($property == null) continue;
            
            // Verificar se esta propriedade tem informao de superpropriedades associada:
            if(count($ontology->_superObjectProperties[$propertyURI]) == 0) continue;
            
            // Percorrer as superpropriedades desta:
            foreach($ontology->_superObjectProperties[$propertyURI] as $superPropertyURI)
            {
                // Efectuar a ligao directa, caso no exista:
                if(!in_array($superPropertyURI, $property->getSuperPropertyURIs()))
                    $property->addSuperPropertyURI($superPropertyURI);
                
                // Obter o objecto da superpropriedade:
                $superProperty = $ontology->getObjectProperty($superPropertyURI);
                if($superProperty == null) continue;
                
                // Efectuar a ligao inversa, caso no exista:
                if(!in_array($propertyURI, $superProperty->getSubPropertyURIs()))
                    $superProperty->addSubPropertyURI($propertyURI);
            }
        }
    }
    
    
    
    /**
     * Efectuar as asseres entre indivduo e propriedades de objecto de uma topologia.
     * @param OWLOntology $ontology Um apontador para a ontologia.
     */
    public function inferOPAssertions(&$ontology)
    {
        // Percorrer todas as propriedades de objecto existentes, caso existam:
        foreach(array_keys($ontology->_OPAssertions) as $individualURI)
        {
            // Obter o objecto do indivduo:
            $individual = $ontology->getIndividual($individualURI);
            if($individual == null) continue;
            
            // Efectuar as asseres necessrias:
            for($i = 0; $i < count($ontology->_OPAssertions[$individualURI]); $i = $i+2)
            {
                $propertyURI = $ontology->_OPAssertions[$individualURI][$i];
                $targetIndividualURI = $ontology->_OPAssertions[$individualURI][$i+1];
                $individual->addOPAssertion($propertyURI, $targetIndividualURI);
            }
        }
    }
    
    
    
    /**
     * Mostra o contedo da topologia no ecr.
     * @param OWLOntology $ontology O apontador para a ontologia a mostrar.
     */
    public function printOntology(&$ontology)
    {
        if(OWLReader::$DEBUG) print "<hr/><br/>";
        
        // Resultado do parsing:
        $error = $this->getErrorDescription();
        if($error != null)
        {
            print "<big><b>AVISO: Erro ao efectuar parsing do ficheiro:</b></big><br/>";
            print utf8_encode("<b>Descrio: </b>") . $error . "<br/>";
            print "<b>Linha: </b>" . $this->getErrorLine() . "<br/>";
            print "<br/>";
            print "<br/>";
        }
        else
        {
            print utf8_encode("<big>Ontologia lida com sucesso!</big>");
            print "<br/>";
            print "<br/>";
        }
        
        // Informaes gerais:
        print utf8_encode("<h1><b>Informaes gerais</b></h1>");
        print "<b>Ficheiro: </b>" . utf8_encode($ontology->getPath()) . "<br/>";
        print "<b>Tipo de estrutura: </b>";
            if(strcmp($ontology->getOntologyType(), OWLTag::$OWL2) == 0)
                print "OWL/XML";
            elseif(strcmp($ontology->getOntologyType(), OWLTag::$OWL) == 0)
                print "RDF/XML";
            else "?";
            print " (" . utf8_encode($ontology->getOntologyType()) . ")<br/>";
        print "<b>Nome interno: </b>" . utf8_encode($ontology->getName()) . "<br/>";
        print "<b>Namespace: </b>" . utf8_encode($ontology->getNamespace()) . "<br/>";
        print utf8_encode("<b>Comentrio: </b>" . $ontology->getComment()) . "<br/>";
        print "<br/>";
        print "<br/>";
        
        // ----------------------------------------------------------------------------------------
        
        // Classes:
        print "<h1><b>Classes</b></h1>";
        print "<ul>";
        foreach($ontology->getClassURIs() as $classURI)
            print "<li><b>" . $ontology->getClass($classURI)->getLabel() . "</b> (" . $classURI . ")</li>";
        print "</ul>";
        print "<b>Total: </b>" . count($ontology->getClassURIs());
        print "<br/>";
        print "<br/>";
        print "<br/>";
        
        // ----------------------------------------------------------------------------------------
        
        // Indivduos:
        print utf8_encode("<h1><b>Indivduos</b></h1>");
        print "<ul>";
        foreach($ontology->getIndividualURIs() as $individualURI)
            print "<li><b>" . $ontology->getIndividual($individualURI)->getLabel() . "</b> (" . $individualURI . ")</li>";
        print "</ul>";
        print "<b>Total: </b>" . count($ontology->getIndividualURIs());
        print "<br/>";
        print "<br/>";
        print "<br/>";
        
        // ----------------------------------------------------------------------------------------
        
        // Propriedades de objecto:
        print utf8_encode("<h1><b>Propriedades de objecto</b></h1>");
        print "<ul>";
        foreach($ontology->getObjectPropertyURIs() as $propertyURI)
            print "<li><b>" . $ontology->getObjectProperty($propertyURI)->getLabel() . "</b> (" . $propertyURI . ")</li>";
        print "</ul>";
        print "<b>Total: </b>" . count($ontology->getObjectPropertyURIs());
        print "<br/>";
        print "<br/>";
        print "<br/>";
        
        // ----------------------------------------------------------------------------------------
        
        // Hierarquia de classes:
        print "<h1><b>Hierarquia de classes</b></h1>";
        print "<ul>";
        OWLReader::aux_hierClassPrint($ontology, $ontology->getThingClassURI());
        print "</ul>";
        print "<br/>";
        
        // ----------------------------------------------------------------------------------------
        
        // Hierarquia de propriedades de objecto:
        print "<h1><b>Hierarquia de propriedades de objecto</b></h1>";
        print "<ul>";
        foreach($ontology->getObjectPropertyURIs() as $propertyURI)
            if(count($ontology->getObjectProperty($propertyURI)->getSuperPropertyURIs()) == 0)
                OWLReader::aux_hierOPPrint($ontology, $propertyURI);
        print "</ul>";
        print "<br/>";
        
        // ----------------------------------------------------------------------------------------
        
        // Membros (indivduos) das classes...
        print "<h1><b>Membros de classes</b></h1>";
        print "<ul>";
        foreach($ontology->getClassURIs() as $classURI)
        {
            $class = $ontology->getClass($classURI);
            if($class == null) continue;
            if(count($class->getMemberURIs()) == 0) continue;
            print "<li><b>" . $class->getLabel() . "</b> tem como membro:</li><ul>";
            foreach($class->getMemberURIs() as $individualURI)
            {
                $individual = $ontology->getIndividual($individualURI);
                if($individual == null) print "<li>" . $individualURI . "</li>";
                else print "<li>" . $individual->getLabel() . "</li>";
            }
            print "</ul>";
        }
        print "</ul>";
        print "<br/>";
        
        // ----------------------------------------------------------------------------------------
        
        // Asseres de propriedades de objecto:
        print utf8_encode("<h1><b>Asseres de propriedades de objecto:</b></h1>");
        print "<ul>";
        foreach($ontology->getIndividualURIs() as $individualURI)
        {
            // Obter o objecto do indivduo:
            $individual = $ontology->getIndividual($individualURI);
            if($individual == null) continue;
            
            // Obter a assero, caso haja:
            $assertions = $individual->getOPAssertionURIs();
            if(count($assertions) == 0) continue;
            
            // Percorrer as asseres:
            foreach($assertions as $assertion)
            {
                // Obter os URIs respectivas:
                $propertyURI = $assertion[0];
                $targetIndividualURI = $assertion[1];
                
                // Obter e mostrar os objectos:
                print "<li>" . $individual->getLabel() . " <i>";
                $property = $ontology->getObjectProperty($propertyURI);
                if($property == null) print "?</i> ";
                else print $property->getLabel() . "</i> ";
                $targetIndividual = $ontology->getIndividual($targetIndividualURI);
                if($targetIndividual == null) print "?";
                else print $targetIndividual->getLabel();
                print "</li>";
            }                 
        }
        print "</ul>";
        print "<br/>";                              
    }
    
    
    
    // Funo auxiliar interna para imprimir a hierarquia de classes.
    private function aux_hierClassPrint(&$ontology, $classURI)
    {
        // Obter o objecto PHP da classe:
        $class = $ontology->getClass($classURI);
        
        // Mostrar dados da classe actual:
        if($class == null)
        { print "<li>?</li>"; return; }
        else
        {
            if(count($class->getSubClassURIs()) == 0)
                print "<li>" . $class->getLabel() . "</li>";
            else print "<li><b>" . $class->getLabel() . "</b></li>";
        }
        
        // Percorrer as subclasses:
        foreach($class->getSubClassURIs() as $nextURI)
        {
            print "<ul>";                                                        
            OWLReader::aux_hierClassPrint($ontology, $nextURI);
            print "</ul>";
        }
    }
    
    
    
    // Funo auxiliar interna para imprimir a hierarquia de propriedades de objecto.
    private function aux_hierOPPrint($ontology, $propertyURI)
    {
        // Obter o objecto PHP da propriedade:
        $property = $ontology->getObjectProperty($propertyURI);
        
        // Mostrar dados da classe actual:
        if($property == null)
        { print "<li>?</li>"; return; }
        else
        {
            if(count($property->getSubPropertyURIs()) == 0)
                print "<li>" . $property->getLabel() . "</li>";
            else print "<li><b>" . $property->getLabel() . "</b></li>";
        }
        
        // Percorrer as subclasses:
        foreach($property->getSubPropertyURIs() as $nextURI)
        {
            print "<ul>";                                                        
            OWLReader::aux_hierOPPrint($ontology, $nextURI);
            print "</ul>";
        }
    }
}
?>